#include <avr/io.h>
#include <inttypes.h>
#include <avr/interrupt.h>
#include <avr/sleep.h>
#include <util/delay.h>
#include <avr/sfr_defs.h>
#include <avr/sleep.h>
#include <avr/eeprom.h> 


//Clock = 1MHz, using the internal RC oscillator 4MHz of mega8
//and dividing by 4 after programming the apropriate fuses


//Data and Code used by ISRs

uint8_t data[] = {0, 0, 0, 0};		//The data to be shown
uint8_t digit = 0;
uint8_t is_cntdwn_started = 0;
uint8_t keep_printing = 1;
uint8_t ovfs = 0;
uint8_t is_cntdwn_over = 0;
uint8_t is_cntdwn_set = 0;
uint16_t seconds = 0;


//uint8_t getButton(volatile uint8_t *in_port, uint8_t pin);
void set_time(uint8_t *data, uint16_t secs);
void sound(uint16_t t);


ISR(BADISR_vect){
}



ISR(INT0_vect)	{						//start or stop cntdwn
	
	PORTB |= 0x0F;
	_delay_ms(100);
	loop_until_bit_is_set(PIND, 2);
	_delay_ms(100);

	if(is_cntdwn_set){
		if(is_cntdwn_started){			
			TCCR2 = 0x00;		
			keep_printing = 0;
	 		//turn off relay
			//go to start again		
		}				
		else{									//1clock = 1.024ms
			
			PORTB &= ~(1<<5);	//turn on relay via optocoupler
			TCCR2 |= (1<<CS20)|(1<<CS21)|(1<<CS22);
			is_cntdwn_started = 1;	
		}
	}
	else{
		keep_printing = 0;	//The first time INT0 is invoked, cntdwn is set via start/stop button
		is_cntdwn_set = 1;	//Cntdwn to be saved in eeprom

	}
}



ISR(INT1_vect){
	
	uint8_t pushDur = 1;

	PORTB |= 0x0F;

	_delay_ms(500);
	if(bit_is_set(PIND, 3)){
		pushDur = 1;
		sound(100);
	}
	else
	{
		loop_until_bit_is_set(PIND, 3);
		pushDur = 2;
		sound(500);
	}
	

	switch(digit){
		case 0:{
			if(pushDur != 2){
				if(data[3] < 9)
					data[3]++;
				else
					data[3] = 0;
				break;
			}
			else
				digit = 1;
			break;
		}
		case 1:{
			if(pushDur != 2){
				if(data[2] < 5)
					data[2] += 1;
				else
					data[2] = 0;
				break;
			}
			else
				digit = 2;
			break;
		}
		case 2:{
			if(pushDur != 2){
				if(data[1] < 9)
					data[1] += 1;
				else
					data[1] = 0;
				break;
			}
			else
				digit = 3;
			break;
		}
		case 3:{
			if(pushDur != 2){
				if(data[0] < 9)
					data[0] += 1;
				else
					data[0] = 0;
				break;
			}
			else{
				//keep_printing = 0;
				digit = 0;
				//is_cntdwn_set = 1;
				//Here countdown is set
				//to be saved in eeprom
				break;
			}
		}
		default:
			break;

	}

}


ISR(TIMER2_OVF_vect){

	ovfs++;
	if(ovfs > 3){
		ovfs = 0;
		seconds--;
		PORTB |= (1<<4);
		if(seconds){
			set_time(data, seconds); //It modifies data
		
		}
		else{
			for(uint8_t i=0; i<4; i++)
				data[i] = 0;
			TCCR2 &= ~((1<<CS20)|(1<<CS21)|(1<<CS22));
			is_cntdwn_over = 1;
			keep_printing = 0;
			//save cntdwn in eeprom
		}
	}
	if(ovfs == 2)
		PORTB &= ~(1<<4);
}

uint8_t bcd(uint8_t dec_num){
	return dec_num&0x0F;
}



//This function gets the N bcd numbers found in data[] array and prints them 
//on the respective N leds.
//To do so, once in every T ms, it energizes every led for ton ms.
//Maximum No of LEDS = 4
//PORTS pins = 4
//Needs keep_printing variable externaly defined
void show_data(volatile uint8_t *led_port, volatile uint8_t *data_port, 
				uint8_t N_leds, uint8_t *data, uint8_t T, uint8_t ton){
	

	
	uint8_t n = N_leds;

	while(keep_printing){
		
		*led_port |= 0x0F;
		*data_port = bcd(*(data + n));
		switch(n){
			case 0:{
				*led_port &= 0b11111110;
				break;
			}
			case 1:{
				*led_port |= 0b00001101;
				*led_port &= ~(1<<1);
				break;
			}
			case 2:{
				*led_port &= 0b11111011;
				break;
			}
			case 3:{
				*led_port &= 0b11110111;
				break;
			}
			default:
				break;
		}

		//start count ton
		_delay_ms(ton);
	
		if(n){
			n--;
		}
		else{
			n = N_leds;
			*led_port |= 0x0F;			//turn off leds
			_delay_ms(T - 4*ton);
					
		}
	}
	*led_port |= 0x0F;
}

void ioinit(void)	{
	


	DDRB = 0x7F;			//portb(0..4) as output to on/off LEDs
	PORTB = 0x7F;
	DDRC = 0x0F;			//portc(0..3) as output BCD number to 7448
	DDRD = 0x00;			//portd(1..3) as input (+ pull up res.) to buttons
	PORTD = 0x0E;

	_delay_ms(1000);		//so that signals in PIND are stabilized
	//set_sleep_mode(SLEEP_MODE_IDLE);
	//sleep_enable();
	

}





/*
uint8_t getButton(volatile uint8_t *in_port, uint8_t pin){
	do{
		loop_until_bit_is_clear(*in_port, pin);		//Wait Until button is pushed	
		_delay_ms(1);									//Get rid of jitter 1ms
		
		if(bit_is_clear(*in_port, pin)){			//If button remains pushed 
			_delay_ms(100);
			if(bit_is_set(*in_port, pin))			//Here after 50ms SET is left, Single push
				return(1);
			else{
				_delay_ms(500);			//Here after 50ms SET remains pushed

				if(bit_is_clear(*in_port, pin)){
					loop_until_bit_is_set(*in_port, pin);		
					return(2);			//Here after 500ms SET is pushed, Permanent push
									//Wait Until button is left
				}
				else
					return(1);
					
			}
				
		}			//here it was jitter

	}while(1);

}*/


void set_time(uint8_t *data, uint16_t secs){
	if(secs>6000){
		for(uint8_t i=0; i<4; i++)
			data[i] = 0;
	}
	else{
		data[0] = (uint8_t)(secs/600);
		secs = secs - data[0]*600;
		data[1] = (uint8_t)(secs/60);
		secs = secs - data[1]*60;
		data[2] = (uint8_t)(secs/10);
		secs = secs - data[2]*10;
		data[3] = (uint8_t)secs;
	
	}

}






void set_cntdwn(void){			//INT0 and INT1 activated
	
	GICR |= (1<<INT1);
	MCUCR |= (1<<ISC11);

	GICR |= (1<<INT0);
	MCUCR |= (1<<ISC01);
	
	sei();
	keep_printing = 1;
	show_data(&PORTB, &PORTC, 3, data, 14, 2);		//show 0 0 0 0 till INT1 appears
	cli();											//or INT0 to accept the cntdwn
													//e.x 34 59 is set

	GICR &= ~(1<<INT1);
	MCUCR &= ~(1<<ISC11);
	
	GICR &= ~(1<<INT0);
	MCUCR &= ~(1<<ISC01);


}													



void get_seconds(uint8_t *data, uint16_t *seconds){

	*seconds = data[3] + data[2]*10 + data[1]*60 + data[0]*600;

}


//Show cntdwn selected until start button pressed
//Only start/stop button INT0 and TOVF2 INT can interrupt show_data()
void start_cntdwn(void){

	//Use TCNT2 to count seconds. 1 sec ~ 4 overflows 1MHz clock/1024

	GICR |= (1<<INT0);
	MCUCR |= (1<<ISC01);
	
	TIMSK |= (1<<TOIE2);



	get_seconds(data, &seconds);	//modifies seconds global variable	
	

	sei();
	keep_printing = 1;	
	show_data(&PORTB, &PORTC, 3, data, 14, 2);	
	cli();

	TIMSK &= ~(1<<TOIE2);

	GICR &= ~(1<<INT0);
	MCUCR &= ~(1<<ISC01);


}

void reinit(void){

	is_cntdwn_started = is_cntdwn_over = is_cntdwn_set = 0;

}

void sound(uint16_t t){

	PORTB &= ~(1<<6);
	_delay_ms(t);
	PORTB |= (1<<6);


}
int main(void)	{
	
	ioinit();	
	while(1){
		eeprom_read_block(data, 0x00, 4); 	//Retrieve last used cntdwn
	
		set_cntdwn();
							//Here cntdwn is set
		_delay_ms(500);		//blink leds and double beep to show acceptance	
		sound(200);
		_delay_ms(200);
		sound(200);


		eeprom_write_block(data, 0x00, 4); 	//Save cntdwn in eeprom

		start_cntdwn();

		//Here cntdwn is over or halted
		//turn off lamps
		PORTB |= (1<<5);
		sound(3000);

		reinit();

	}

	
	return(0);

}
